@@ -14,8 +14,8 @@ Other potentially useful links:
 """
 
 from base64 import b32encode
-from collections import namedtuple
-from datetime import date, datetime, timezone
+from datetime import date
+from enum import Enum
 import hashlib
 import logging
 import re
@@ -27,17 +27,18 @@ from requests.exceptions import (ChunkedEncodingError,
                                  RetryError,
                                  Timeout)
 import time
-import urllib.parse
 from urllib3.connectionpool import HTTPConnectionPool
 from urllib3.exceptions import (ConnectTimeoutError,
                                 MaxRetryError,
                                 ReadTimeoutError)
 from . import _utils, __version__
+from ._models import CdxRecord, Memento
 from .exceptions import (WaybackException,
                          UnexpectedResponseFormat,
                          BlockedByRobotsError,
                          BlockedSiteError,
                          MementoPlaybackError,
+                         NoMementoError,
                          WaybackRetryError,
                          RateLimitError)
 
@@ -54,11 +55,7 @@ CDX_SEARCH_URL = 'http://web.archive.org/cdx/search/cdx'
 # TODO: support new/upcoming CDX API
 # CDX_SEARCH_URL = 'http://web.archive.org/web/timemap/cdx'
 
-ARCHIVE_RAW_URL_TEMPLATE = 'http://web.archive.org/web/{timestamp}id_/{url}'
-ARCHIVE_VIEW_URL_TEMPLATE = 'http://web.archive.org/web/{timestamp}/{url}'
-URL_DATE_FORMAT = '%Y%m%d%H%M%S'
-MEMENTO_URL_PATTERN = re.compile(
-    r'^http(?:s)?://web.archive.org/web/(\d+)(?:id_)?/(.+)$')
+ARCHIVE_URL_TEMPLATE = 'http://web.archive.org/web/{timestamp}{mode}/{url}'
 REDUNDANT_HTTP_PORT = re.compile(r'^(http://[^:/]+):80(.*)$')
 REDUNDANT_HTTPS_PORT = re.compile(r'^(https://[^:/]+):443(.*)$')
 DATA_URL_START = re.compile(r'data:[\w]+/[\w]+;base64')
@@ -72,129 +69,58 @@ EMAILISH_URL = re.compile(r'^https?://(<*)((mailto:)|([^/@:]*@))')
 # Make sure it roughly starts with a valid protocol + domain + port?
 URL_ISH = re.compile(r'^[\w+\-]+://[^/?=&]+\.\w\w+(:\d+)?(/|$)')
 
-CdxRecord = namedtuple('CdxRecord', (
-    # Raw CDX values
-    'key',
-    'timestamp',
-    'url',
-    'mime_type',
-    'status_code',
-    'digest',
-    'length',
-    # Synthesized values
-    'raw_url',
-    'view_url'
-))
-"""
-Item from iterable of results returned by :meth:`WaybackClient.search`
 
-These attributes contain information provided directly by CDX.
-
-.. py:attribute:: digest
-
-   Content hashed as a base 32 encoded SHA-1.
-
-.. py:attribute:: key
-
-   SURT-formatted URL
-
-.. py:attribute:: length
-
-   Size of captured content in bytes, such as :data:`2767`. This may be
-   innacurate. If the record is a "revisit record", indicated by MIME type
-   :data:`'warc/revisit'`, the length seems to be the length of the reference,
-   not the length of the content itself.
-
-.. py:attribute:: mime_type
-
-   MIME type of record, such as :data:`'text/html'`, :data:`'warc/revisit'` or
-   :data:`'unk'` ("unknown") if this information was not captured.
-
-.. py:attribute:: status_code
-
-   Status code returned by the server when the record was captured, such as
-   :data:`200`. This is may be :data:`None` if the record is a revisit record.
-
-.. py:attribute:: timestamp
-
-   The capture time represented as a :class:`datetime.datetime`, such as
-   :data:`datetime.datetime(1996, 12, 31, 23, 58, 47)`.
-
-.. py:attribute:: url
-
-   The URL that was captured by this record, such as
-   :data:`'http://www.nasa.gov/'`.
-
-And these attributes are synthesized from the information provided by CDX.
-
-.. py:attribute:: raw_url
-
-   The URL to the raw captured content, such as
-   :data:`'http://web.archive.org/web/19961231235847id_/http://www.nasa.gov/'`.
-
-.. py:attribute:: view_url
-
-   The URL to the public view on Wayback Machine. In this view, the links and
-   some subresources in the document are rewritten to point to Wayback URLs.
-   There is also a navigation panel around the content. Example URL:
-   :data:`'http://web.archive.org/web/19961231235847/http://www.nasa.gov/'`.
-"""
-
-
-def split_memento_url(memento_url):
-    'Extract the raw date and URL components from a memento URL.'
-    match = MEMENTO_URL_PATTERN.match(memento_url)
-    if match is None:
-        raise ValueError(f'"{memento_url}" is not a memento URL')
-
-    return match.group(2), match.group(1)
-
-
-def clean_memento_url_component(url):
-    # A URL *may* be percent encoded, decode ONLY if so (we don’t want to
-    # accidentally decode the querystring if there is one)
-    lower_url = url.lower()
-    if lower_url.startswith('http%3a') or lower_url.startswith('https%3a'):
-        url = urllib.parse.unquote(url)
-
-    return url
-
-
-def memento_url_data(memento_url):
+class Mode(Enum):
     """
-    Get the original URL and date that a memento URL represents a capture of.
+    An enum describing the playback mode of a memento. When requesting a
+    memento (e.g. with :meth:`wayback.WaybackClient.get_memento`), you can use
+    these values to determine how the response body should be formatted.
+
+    For more details, see:
+    http://archive-access.sourceforge.net/projects/wayback/administrator_manual.html#Archival_URL_Replay_Mode
 
     Examples
     --------
-    Extract original URL and date.
+    >>> waybackClient.get_memento('https://noaa.gov/',
+    >>>                           datetime=datetime.datetime(2018, 1, 2),
+    >>>                           mode=wayback.Mode.view)
 
-    >>> url = ('http://web.archive.org/web/20170813195036/'
-    ...        'https://arpa-e.energy.gov/?q=engage/events-workshops')
-    >>> memento_url_data(url)
-    ('https://arpa-e.energy.gov/?q=engage/events-workshops',
-     datetime.datetime(2017, 8, 13, 19, 50, 36))
+    **Values**
+
+    .. py:attribute:: original
+
+        Returns the HTTP response body as originally captured.
+
+    .. py:attribute:: view
+
+        Formats the response body so it can be viewed with a web
+        browser. URLs for links and subresources like scripts, stylesheets,
+        images, etc. will be modified to point to the equivalent memento in the
+        Wayback Machine so that the resulting page looks as similar as possible
+        to how it would have appeared when originally captured. It's mainly meant
+        for use with HTML pages. This is the playback mode you typically use when
+        browsing the Wayback Machine with a web browser.
+
+    .. py:attribute:: javascript
+
+        Formats the response body by updating URLs, similar
+        to ``Mode.view``, but designed for JavaScript instead of HTML.
+
+    .. py:attribute:: css
+
+        Formats the response body by updating URLs, similar to
+        ``Mode.view``, but designed for CSS instead of HTML.
+
+    .. py:attribute:: image
+
+        formats the response body similar to ``Mode.view``, but
+        designed for image files instead of HTML.
     """
-    raw_url, timestamp = split_memento_url(memento_url)
-    url = clean_memento_url_component(raw_url)
-    date = datetime.strptime(timestamp, URL_DATE_FORMAT)
-
-    return url, date
-
-
-def original_url_for_memento(memento_url):
-    """
-    Get the original URL that a memento URL represents a capture of.
-
-    Examples
-    --------
-    Extract original URL.
-
-    >>> url = ('http://web.archive.org/web/20170813195036/'
-    ...        'https://arpa-e.energy.gov/?q=engage/events-workshops')
-    >>> original_url_for_memento(url)
-    'https://arpa-e.energy.gov/?q=engage/events-workshops'
-    """
-    return clean_memento_url_component(split_memento_url(memento_url)[0])
+    original = 'id_'
+    view = ''
+    javascript = 'js_'
+    css = 'cs_'
+    image = 'im_'
 
 
 def is_malformed_url(url):
@@ -283,8 +209,8 @@ class WaybackSession(_utils.DisableAfterCloseSession, requests.Session):
         So if this was `4`, retries would happen after the following delays:
         0 seconds, 4 seconds, 8 seconds, 16 seconds, ...
     timeout : int or float or tuple of (int or float, int or float), optional
-        A timeout to use for all requests. If not set, there will be no
-        no explicit timeout. See the Requests docs for more:
+        A timeout to use for all requests. (Default: ``60``)
+        See the Requests docs for more:
         http://docs.python-requests.org/en/master/user/advanced/#timeouts
     user_agent : str, optional
         A custom user-agent string to use in all requests. Defaults to:
@@ -301,7 +227,7 @@ class WaybackSession(_utils.DisableAfterCloseSession, requests.Session):
     # just the error type. See `should_retry_error()`.
     handleable_errors = (ConnectionError,) + retryable_errors
 
-    def __init__(self, retries=6, backoff=2, timeout=None, user_agent=None):
+    def __init__(self, retries=6, backoff=2, timeout=60, user_agent=None):
         super().__init__()
         self.retries = retries
         self.backoff = backoff
@@ -563,16 +489,8 @@ class WaybackClient(_utils.DepthCountedContext):
             if value is not None:
                 if isinstance(value, str):
                     final_query[key] = value
-                elif isinstance(value, datetime):
-                    # Make sure we have either a naive datetime (assumed to
-                    # represent UTC) or convert the datetime to UTC.
-                    if value.tzinfo:
-                        value_utc = value.astimezone(timezone.utc)
-                    else:
-                        value_utc = value
-                    final_query[key] = value_utc.strftime(URL_DATE_FORMAT)
                 elif isinstance(value, date):
-                    final_query[key] = value.strftime(URL_DATE_FORMAT)
+                    final_query[key] = _utils.format_timestamp(value)
                 else:
                     final_query[key] = str(value).lower()
 
@@ -622,9 +540,7 @@ class WaybackClient(_utils.DepthCountedContext):
                 else:
                     status_code = int(data.status_code)
                 length = int(data.length)
-                capture_time = (datetime.strptime(data.timestamp,
-                                                  URL_DATE_FORMAT)
-                                        .replace(tzinfo=timezone.utc))
+                capture_time = _utils.parse_timestamp(data.timestamp)
             except Exception as err:
                 if 'RobotAccessControlException' in text:
                     raise BlockedByRobotsError(query["url"])
@@ -647,54 +563,90 @@ class WaybackClient(_utils.DepthCountedContext):
                 status_code=status_code,
                 length=length,
                 timestamp=capture_time,
-                raw_url=ARCHIVE_RAW_URL_TEMPLATE.format(
-                    timestamp=data.timestamp, url=data.url),
-                view_url=ARCHIVE_VIEW_URL_TEMPLATE.format(
-                    timestamp=data.timestamp, url=data.url)
+                raw_url=ARCHIVE_URL_TEMPLATE.format(
+                    url=data.url,
+                    timestamp=data.timestamp,
+                    mode=Mode.original.value),
+                view_url=ARCHIVE_URL_TEMPLATE.format(
+                    url=data.url,
+                    timestamp=data.timestamp,
+                    mode=Mode.view.value)
             )
             count += 1
             yield data
 
         return count
 
-    # TODO: make this nicer by taking an optional date, so `url` can be a
-    # memento url or an original URL + plus date and we'll compose a memento
-    # URL.
-    # TODO: for generic use, needs to be able to return the memento itself if
-    # the memento was a redirect (different than allowing a nearby-in-time
-    # memento of the same URL, which would be the above argument). Probably
-    # call this `follow_redirects=True`?
-    def get_memento(self, url, exact=True, exact_redirects=None,
-                    target_window=24 * 60 * 60):
+    def get_memento(self, url, datetime=None, mode=Mode.original, *,
+                    exact=True, exact_redirects=None,
+                    target_window=24 * 60 * 60, follow_redirects=True):
         """
-        Fetch a memento from the Wayback Machine. This retrieves the content
-        that was ultimately returned from a memento, following any redirects
-        that were present at the time the memento was captured. (That is, if
-        `http://example.com/a` redirected to `http://example.com/b`, this
-        returns the memento for `/b` when you request `/a`.)
+        Fetch a memento (an archived HTTP response) from the Wayback Machine.
+
+        Not all mementos can be successfully fetched (or “played back” in
+        Wayback terms). In this case, ``get_memento`` can load the
+        next-closest-in-time memento or it will raise
+        :class:`wayback.exceptions.MementoPlaybackError` depending on the value
+        of the ``exact`` and ``exact_redirects`` parameters (see more details
+        below).
 
         Parameters
         ----------
-        url : string
-            URL of memento in Wayback (e.g.
-            `http://web.archive.org/web/20180816111911id_/http://www.nws.noaa.gov/sp/`)
+        url : string or CdxRecord
+            URL to retrieve a memento of. This can be any of:
+
+            - A normal URL (e.g. ``http://www.noaa.gov/``). When using this
+              form, you must also specify ``datetime``.
+            - A ``CdxRecord`` retrieved from
+              :meth:`wayback.WaybackClient.search`.
+            - A URL of the memento in Wayback, e.g.
+              ``http://web.archive.org/web/20180816111911id_/http://www.noaa.gov/``
+
+        datetime : datetime.datetime or datetime.date or str, optional
+            The time at which to retrieve a memento of ``url``. If ``url`` is
+            a :class:`wayback.CdxRecord` or full memento URL, this parameter
+            can be omitted.
+        mode : wayback.Mode or str, optional
+            The playback mode of the memento. This determines whether the
+            content of the returned memento is exactly as originally captured
+            (the default) or modified in some way. See :class:`wayback.Mode`
+            for a description of possible values.
+
+            For more details, see:
+            http://archive-access.sourceforge.net/projects/wayback/administrator_manual.html#Archival_URL_Replay_Mode
+
+            Default: :py:attr:`wayback.Mode.original`
+
         exact : boolean, optional
             If false and the requested memento either doesn't exist or can't be
             played back, this returns the closest-in-time memento to the
-            requested one, so long as it is within `target_window`.
+            requested one, so long as it is within ``target_window``. If there
+            was no memento in the target window or if ``exact=True``, then this
+            will raise :class:`wayback.exceptions.MementoPlaybackError`.
             Default: True
         exact_redirects : boolean, optional
             If false and the requested memento is a redirect whose *target*
-            doesn't exist or or can't be played back, this returns the closest-
-            in-time memento to the intended target, so long as it is within
-            `target_window`. If unset, this will be the same as `exact`.
+            doesn't exist or can't be played back, this returns the
+            closest-in-time memento to the intended target, so long as it is
+            within ``target_window``. If unset, this will be the same as
+            ``exact``.
         target_window : int, optional
             If the memento is of a redirect, allow up to this many seconds
-            between the capture of the redirect and the capture of the target
-            URL. (Note this does NOT apply when the originally requested
-            memento didn't exist and wayback redirects to the next-closest-in-
-            -time one. That will always raise a MementoPlaybackError.)
+            between the capture of the redirect and the capture of the
+            redirect's target URL. This window also applies to the first
+            memento if ``exact=False`` and the originally
+            requested memento was not available.
             Defaults to 86,400 (24 hours).
+        follow_redirects : boolean, optional
+            If true (the default), ``get_memento`` will follow historical
+            redirects to return the content that a web browser would have
+            ultimately displayed at the requested URL and time, rather than the
+            memento of an HTTP redirect response (i.e. a 3xx status code).
+            That is, if ``http://example.com/a`` redirected to
+            ``http://example.com/b``, then this method returns the memento for
+            ``/a`` when ``follow_redirects=False`` and the memento for ``/b``
+            when ``follow_redirects=True``.
+            Default: True
 
         Returns
         -------
@@ -702,11 +654,44 @@ class WaybackClient(_utils.DepthCountedContext):
             An HTTP response with the content of the memento, including a
             history of any redirects involved. (For a complete history of all
             HTTP requests needed to obtain the memento [rather than historic
-            redirects], check `debug_history` instead of `history`.)
+            redirects], check ``debug_history`` instead of ``history``.)
         """
         if exact_redirects is None:
             exact_redirects = exact
 
+        # Convert Mode enum values to strings rather than the other way around.
+        # There is a very real possibility of other undocumented modes and we
+        # don't want converting them via `Mode(str_value)` to raise an error
+        # and make them impossible to use. Wayback folks have been unclear
+        # about exactly what all the options are, and it sounds kind of like
+        # this is an old and kind of messy part of the codebase, where things
+        # may have been changed around a few times and the people who worked on
+        # it are now off doing other things.
+        if isinstance(mode, Mode):
+            mode = mode.value
+        elif not isinstance(mode, str):
+            raise TypeError('`mode` must be a wayback.Mode or string '
+                            f'(received {mode!r})')
+
+        if isinstance(url, CdxRecord):
+            original_date = url.timestamp
+            original_url = url.url
+        else:
+            try:
+                original_url, original_date, mode = _utils.memento_url_data(url)
+            except ValueError:
+                original_url = url
+                if not datetime:
+                    raise TypeError('You must specify `datetime` when using a '
+                                    'normal URL for get_memento()')
+                else:
+                    original_date = _utils.ensure_utc_datetime(datetime)
+
+        original_date_wayback = _utils.format_timestamp(original_date)
+        url = ARCHIVE_URL_TEMPLATE.format(timestamp=original_date_wayback,
+                                          mode=mode,
+                                          url=original_url)
+
         with _utils.rate_limited(calls_per_second=30, group='get_memento'):
             # Correctly following redirects is actually pretty complicated. In
             # the simplest case, a memento is a simple web page, and that's
@@ -737,13 +722,30 @@ class WaybackClient(_utils.DepthCountedContext):
             debug_history = []
             urls = set()
             previous_was_memento = False
-            orginal_url, original_date = memento_url_data(url)
+
             response = self.session.request('GET', url, allow_redirects=False)
             protocol_and_www = re.compile(r'^https?://(www\d?\.)?')
+            memento = None
             while True:
                 is_memento = 'Memento-Datetime' in response.headers
+                current_url, current_date, current_mode = _utils.memento_url_data(response.url)
 
-                if not is_memento:
+                if is_memento:
+                    memento = Memento(url=current_url,
+                                      timestamp=current_date,
+                                      mode=current_mode,
+                                      memento_url=response.url,
+                                      status_code=response.status_code,
+                                      headers=Memento.parse_memento_headers(response.headers, response.url),
+                                      encoding=response.encoding,
+                                      raw=response,
+                                      raw_headers=response.headers,
+                                      history=history,
+                                      debug_history=debug_history)
+                    if not follow_redirects:
+                        break
+                else:
+                    memento = None
                     # The exactness requirements for redirects from memento
                     # playbacks and non-playbacks is different -- even with
                     # strict matching, a memento that redirects to a non-
@@ -754,8 +756,7 @@ class WaybackClient(_utils.DepthCountedContext):
                     if response.next and (
                        (len(history) == 0 and not exact) or
                        (len(history) > 0 and (previous_was_memento or not exact_redirects))):
-                        current_url = original_url_for_memento(response.url)
-                        target_url, target_date = memento_url_data(response.next.url)
+                        target_url, target_date, _ = _utils.memento_url_data(response.next.url)
                         # A non-memento redirect is generally taking us to the
                         # closest-in-time capture of the same URL. Note that is
                         # NOT the next capture -- i.e. the one that would have
@@ -799,9 +800,19 @@ class WaybackClient(_utils.DepthCountedContext):
                         elif message:
                             raise MementoPlaybackError(f'Memento at {url} could not be played: {message}')
                         elif response.ok:
+                            # TODO: Raise more specific errors for the possible
+                            # cases here. We *should* only arrive here when
+                            # there's a redirect and:
+                            # - `exact` is true.
+                            # - `exact_redirects` is true and the redirect was
+                            #   not exact.
+                            # - The target URL is outside `target_window`.
                             raise MementoPlaybackError(f'Memento at {url} could not be played')
+                        elif response.status_code == 404:
+                            raise NoMementoError(f'The URL {url} has no mementos and was never archived')
                         else:
-                            response.raise_for_status()
+                            raise MementoPlaybackError(f'{response.status_code} error while loading '
+                                                       f'memento at {url}')
 
                 if response.next:
                     previous_was_memento = is_memento
@@ -814,13 +825,11 @@ class WaybackClient(_utils.DepthCountedContext):
 
                     # All requests are included in `debug_history`, but
                     # `history` only shows redirects that were mementos.
-                    debug_history.append(response)
+                    debug_history.append(response.url)
                     if is_memento:
-                        history.append(response)
+                        history.append(memento)
                     response = self.session.send(response.next, allow_redirects=False)
                 else:
                     break
 
-            response.history = history
-            response.debug_history = debug_history
-            return response
+            return memento
